from __future__ import annotations

from enum import Enum

from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.wikidata import ExtendedWikidataNamespace


class KGName(str, Enum):
    Wikidata = "wikidata"
    Generic = "generic"


registered_kgns: dict[str, KnowledgeGraphNamespace] = {
    KGName.Wikidata: ExtendedWikidataNamespace.create(),
}


def get_kgns(kgname: KGName) -> KnowledgeGraphNamespace:
    if kgname in registered_kgns:
        return registered_kgns[kgname]
    raise NotImplementedError(kgname)


def register_kgns(kgname: str, kgns: KnowledgeGraphNamespace):
    global registered_kgns
    registered_kgns[kgname] = kgns


def has_kgns(kgname: str) -> bool:
    global registered_kgns
    return kgname in registered_kgns



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping
from typing import List, Optional, Tuple, Union


class PrefixIndex:
    """Namespace indexing so we can quickly get prefix of a URI."""

    __slots__ = ("index", "start", "end")

    def __init__(
        self, index: dict[str, PrefixIndex | str], start: int, end: int
    ) -> None:
        self.index = index
        self.start = start
        self.end = end

    @staticmethod
    def create(ns2prefix: Mapping[str, str]):
        sorted_ns = sorted(ns2prefix.keys(), key=lambda x: len(x), reverse=True)
        if len(sorted_ns) == 0:
            raise Exception("No namespace provided")

        return PrefixIndex._create(ns2prefix, sorted_ns, 0)

    @staticmethod
    def _create(ns2prefix: Mapping[str, str], nses: List[str], start: int):
        shortest_ns = nses[-1]
        index = PrefixIndex({}, start, len(shortest_ns))

        if index.start == index.end:
            # we have an empty key, it must have more than one element because of the previous call
            index.index[""] = ns2prefix[nses[-1]]
            subindex = PrefixIndex._create(ns2prefix, nses[:-1], index.end)
            for key, node in subindex.index.items():
                index.index[key] = node
            return index

        tmp = defaultdict(list)
        for ns in nses:
            key = ns[index.start : index.end]
            tmp[key].append(ns)

        for key, lst_ns in tmp.items():
            if len(lst_ns) == 1:
                index.index[key] = ns2prefix[lst_ns[0]]
            else:
                index.index[key] = PrefixIndex._create(ns2prefix, lst_ns, index.end)
        return index

    def get(self, uri: str) -> Optional[str]:
        """Get prefix of an uri. Return None if it is not found"""
        key = uri[self.start : self.end]
        if key in self.index:
            value = self.index[key]
            if isinstance(value, PrefixIndex):
                return value.get(uri)
            return value

        if "" in self.index:
            return self.index[""]  # type: ignore

        return None

    def __str__(self):
        """Readable version of the index"""
        stack: List[Tuple[int, str, Union[str, PrefixIndex]]] = list(
            reversed([(0, k, v) for k, v in self.index.items()])
        )
        out = []

        while len(stack) > 0:
            depth, key, value = stack.pop()
            indent = "    " * depth
            if isinstance(value, str):
                out.append(indent + "`" + key + "`: " + value + "\n")
            else:
                out.append(indent + "`" + key + "`:" + "\n")
                for k, v in value.index.items():
                    stack.append((depth + 1, k, v))

        return "".join(out)

    def to_dict(self):
        return {
            k: v.to_dict() if isinstance(v, PrefixIndex) else v
            for k, v in self.index.items()
        }



from __future__ import annotations

from collections import OrderedDict, defaultdict
from typing import List, Optional, Sequence, Tuple, Union

import pandas as pd
from sm.inputs.column import Column


class ColumnBasedTable:
    __slots__ = ("table_id", "columns", "index2columns", "_df")

    def __init__(self, table_id: str, columns: List[Column]):
        self.table_id = table_id
        self.columns = columns
        self.index2columns = {col.index: col for col in columns}
        self._df: Optional[pd.DataFrame] = None

    @property
    def df(self) -> pd.DataFrame:
        if self._df is None:
            self._df = self.as_dataframe()
        return self._df

    def keep_columns(self, columns: List[int], reindex: bool = False):
        if reindex:
            newcols = []
            for i, c in enumerate(columns):
                tmp = self.index2columns[c]
                newcols.append(Column(i, tmp.name, tmp.values))
        else:
            newcols = [self.index2columns[c] for c in columns]
        return ColumnBasedTable(self.table_id, newcols)

    def shape(self) -> Tuple[int, int]:
        """Get shape of table: (number of rows, number of columns)"""
        if len(self.columns) == 0:
            return 0, 0
        return len(self.columns[0].values), len(self.columns)

    def select_rows(self, indices: list[int]) -> ColumnBasedTable:
        """Select a subset of rows based on a boolean mask"""
        return ColumnBasedTable(
            self.table_id, [col.select_rows(indices) for col in self.columns]
        )

    def ncols(self) -> int:
        return len(self.columns)

    def nrows(self) -> int:
        if len(self.columns) == 0:
            return 0
        return len(self.columns[0].values)

    def get_column_by_index(self, col_idx: int) -> Column:
        return self.index2columns[col_idx]

    def as_dataframe(self) -> pd.DataFrame:
        d = OrderedDict()
        dup = defaultdict(int)
        for i, col in enumerate(self.columns):
            if col.name is None:
                cname = f"unk_{i:02}"
            else:
                cname = col.name

            if cname in d:
                dup[cname] += 1
                cname += f" ({dup[cname]})"
            d[cname] = col.values
        return pd.DataFrame(d)

    def subset(self, start_row: int, end_row: int):
        return ColumnBasedTable(
            self.table_id,
            [
                Column(c.index, c.name, c.values[start_row:end_row])
                for c in self.columns
            ],
        )

    def to_dict(self):
        return {
            "version": 2,
            "table_id": self.table_id,
            "columns": [col.to_dict() for col in self.columns],
        }

    def __getitem__(self, item: Tuple[int, int]):
        return self.columns[item[1]][item[0]]

    @staticmethod
    def from_dict(record: dict):
        version = record.get("version", None)
        assert version == 2 or version == "2", version
        return ColumnBasedTable(
            record["table_id"],
            [
                Column(col["index"], col["name"], col["values"])
                for col in record["columns"]
            ],
        )

    @staticmethod
    def from_dataframe(df: pd.DataFrame, table_id: str):
        columns = []
        for ci, c in enumerate(df.columns):
            values = [r.iloc[ci] for ri, r in df.iterrows()]
            column = Column(ci, c, values)
            columns.append(column)
        return ColumnBasedTable(table_id, columns)

    @staticmethod
    def from_rows(
        records: Sequence[Sequence[Union[str, int, float, bool]]],
        table_id: str,
        headers: Optional[Sequence[str]] = None,
        strict: bool = False,
    ):
        if len(records) == 0:
            return ColumnBasedTable(table_id, [])

        if headers is None:
            headers = ["" for _ in range(len(records[0]))]

        if strict:
            ncols = len(headers)
            assert all(
                len(r) == ncols for r in records
            ), "All rows must have the same number of columns"

        columns = []
        for ci, col in enumerate(headers):
            columns.append(Column(ci, col, values=[r[ci] for r in records]))

        return ColumnBasedTable(table_id, columns)



from __future__ import annotations

import re

from rdflib import RDFS
from sm.namespaces.namespace import KnowledgeGraphNamespace, OutOfNamespace
from sm.namespaces.prefix_index import PrefixIndex


class WikidataNamespace(KnowledgeGraphNamespace):
    """Namespace for Wikidata entities and ontology.

    In Wikidata, everything is an entity (classes, properties, items, etc.). But they also have decicated namespaces for their ontology predicates (/prop/).
    So for semantic models, we use the namespace for their properties instead of treating them as entities.
    For example, we use p:P131 instead of wd:P131.
    """

    __slots__ = ("entity_prefix", "entity_ns", "property_prefix", "property_ns")

    # the root entity of Wikidata, any other class is a subclass of this
    entity_id: str = "Q35120"
    entity_uri: str = "http://www.wikidata.org/entity/Q35120"
    entity_label: str = "Entity (Q35120)"  # the label of the root entity
    # statement to represent n-ary relations
    statement_uri: str = "http://wikiba.se/ontology#Statement"
    main_namespaces: list[str] = [
        "http://www.wikidata.org/prop/",
        "http://www.wikidata.org/entity/",
    ]

    URI_RE = re.compile(
        r"^https?:\/\/www\.wikidata\.org\/(?:entity\/|prop\/|wiki\/Property:|wiki\/)([QPL]\d+)$"
    )
    IS_WIKIDATA_URI = re.compile(r"^https?:\/\/www\.wikidata\.org\/")

    @classmethod
    def create(cls):
        prefix2ns = {
            "p": "http://www.wikidata.org/prop/",
            "pq": "http://www.wikidata.org/prop/qualifier/",
            "pqn": "http://www.wikidata.org/prop/qualifier/value-normalized/",
            "pqv": "http://www.wikidata.org/prop/qualifier/value/",
            "pr": "http://www.wikidata.org/prop/reference/",
            "prn": "http://www.wikidata.org/prop/reference/value-normalized/",
            "prv": "http://www.wikidata.org/prop/reference/value/",
            "ps": "http://www.wikidata.org/prop/statement/",
            "psn": "http://www.wikidata.org/prop/statement/value-normalized/",
            "psv": "http://www.wikidata.org/prop/statement/value/",
            "wd": "http://www.wikidata.org/entity/",
            "wdata": "http://www.wikidata.org/wiki/Special:EntityData/",
            "wdno": "http://www.wikidata.org/prop/novalue/",
            "wdref": "http://www.wikidata.org/reference/",
            "wds": "http://www.wikidata.org/entity/statement/",
            "wdt": "http://www.wikidata.org/prop/direct/",
            "wdtn": "http://www.wikidata.org/prop/direct-normalized/",
            "wdv": "http://www.wikidata.org/value/",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "wikibase": "http://wikiba.se/ontology#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        }
        ns2prefix = {v: k for k, v in prefix2ns.items()}
        assert len(ns2prefix) == len(prefix2ns), "Duplicated namespaces"
        prefix_index = PrefixIndex.create(ns2prefix)

        return cls(prefix2ns, ns2prefix, prefix_index)

    def __init__(
        self,
        prefix2ns: dict[str, str],
        ns2prefix: dict[str, str],
        prefix_index: PrefixIndex,
    ):
        super().__init__(prefix2ns, ns2prefix, prefix_index)
        self.entity_prefix = "wd"
        self.property_prefix = "p"
        self.entity_ns = self.prefix2ns[self.entity_prefix]
        self.property_ns = self.prefix2ns[self.property_prefix]

    ###############################################################################
    # URI testing
    ###############################################################################

    @classmethod
    def is_abs_uri_statement(cls, uri: str):
        return uri == WikidataNamespace.statement_uri

    @classmethod
    def is_abs_uri_property(cls, uri: str):
        return uri.startswith(f"http://www.wikidata.org/prop/P") or uri.startswith(
            f"http://www.wikidata.org/entity/P"
        )

    @classmethod
    def is_abs_uri_qnode(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/Q")

    @classmethod
    def is_abs_uri_lexeme(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/L")

    @classmethod
    def is_abs_uri_entity(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/")

    ###############################################################################
    # Converting between URI and ID
    ###############################################################################

    def is_id(self, id: str):
        return (id[0] == "Q" or id[0] == "P" or id[0] == "L") and id[1:].isdigit()

    def is_uri_in_main_ns(self, uri: str) -> bool:
        return self.URI_RE.match(uri) is not None

    def uri_to_id(self, uri: str) -> str:
        m = self.URI_RE.match(uri)
        if m is None:
            raise OutOfNamespace(f"{uri} is not in wikidata namespace")
        return m.group(1)

    def id_to_uri(self, id: str):
        assert self.is_id(id), id
        if id[0] == "Q" or id[0] == "L":
            return f"{self.entity_ns}{id}"
        assert id[0] == "P"
        return f"{self.property_ns}{id}"

    def has_encrypted_name(self, uri: str):
        return self.IS_WIKIDATA_URI.match(uri) is not None


class ExtendedWikidataNamespace(WikidataNamespace):
    extra_properties = {str(RDFS.label)}

    def uri_to_id(self, uri: str) -> str:
        if uri in self.extra_properties:
            return uri
        return super().uri_to_id(uri)

    def id_to_uri(self, id: str):
        if id in self.extra_properties:
            return id
        return super().id_to_uri(id)
